# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import numpy as np

from gwbench.utils import time_fac, REarth, halfTEarth, cLight, is_subset_lists, log_msg

cos = np.cos
sin = np.sin
exp = np.exp

ap_symbs_string = 'f Mc tc ra dec psi gmst0'

def detector_response(f, hfp, hfc, Mc, tc, ra, dec, psi, gmst0, loc, use_rot, user_locs=None):
    # input:    f       frequency domain [Hz]
    #           hfc     cross polarization
    #           hfp     plus polarization
    #           Mc      chirp Mass [solar mass]
    #           tc      time of coalescence [s]
    #           dec     declination [rad]
    #           ra      right ascencsion [rad]
    #           psi     polarization angle [rad]
    #           gmst0   GreenwichMeanSiderialTime according to LAL
    #           loc     location (and implied orientation) of a detector
    #           use_rot use frequency dependent time due to rotation of earth and SPA
    #
    # output:   hf      detector strain

    Fp, Fc, Flp = antenna_pattern_and_loc_phase_fac(f, Mc, tc, ra, dec, psi, gmst0, loc, use_rot, user_locs=user_locs)
    return Flp * (Fp * hfp + Fc * hfc)

def antenna_pattern_and_loc_phase_fac(f, Mc, tc, ra, dec, psi, gmst0, loc, use_rot, user_locs=None):
    # input:    f       frequency domain [Hz]
    #           Mc      chirp Mass [solar mass]
    #           tc      time of coalescence [s]
    #           dec     declination [rad]
    #           ra      right ascencsion [rad]
    #           psi     polarization angle [rad]
    #           gmst0   GreenwichMeanSiderialTime according to LAL
    #           loc     location (and implied orientation) of a detector
    #           use_rot use frequency dependent time due to rotation of earth and SPA
    #
    # output:   Fp, Fc, Flp

    gra  = calc_gra(ra, gmst0, use_rot, f, Mc, tc)
    D, d = det_ten_and_loc_vec(loc, REarth, user_locs=user_locs)

    return *ant_pat_funcs(D, *ant_pat_vectors(gra, dec, psi)), loc_phase_func(gra, dec, f, d)

def calc_gra(ra, gmst0, use_rot, f, Mc, tc):
    if use_rot: return gmst0 - ra + np.pi / halfTEarth * (tc - 5. / 256 * (time_fac * Mc)**(-5./3) * (np.pi * f)**(-8./3))
    else:       return np.array([gmst0 - ra])

def loc_phase_func(gra, dec, f, d):
    theta = np.pi/2 - dec
    return exp(1j * 2 * np.pi * f * np.matmul(d, np.array([cos(gra)*sin(theta), sin(gra)*sin(theta), cos(theta)*np.ones_like(gra)])))

def ant_pat_funcs(D, XX, YY):
    return (0.5 * (np.matmul(D,XX) * XX - np.matmul(D,YY) * YY)).sum(axis=0), \
           (0.5 * (np.matmul(D,XX) * YY + np.matmul(D,YY) * XX)).sum(axis=0)

def ant_pat_vectors(gra, dec, psi):
    return np.array([ -cos(psi)*sin(gra) - sin(psi)*cos(gra)*sin(dec),
                      -cos(psi)*cos(gra) + sin(psi)*sin(gra)*sin(dec),
                                np.ones_like(gra) * sin(psi)*cos(dec) ]), \
           np.array([  sin(psi)*sin(gra) - cos(psi)*cos(gra)*sin(dec),
                       sin(psi)*cos(gra) + cos(psi)*sin(gra)*sin(dec),
                                np.ones_like(gra) * cos(psi)*cos(dec) ])

def det_ten_and_loc_vec(loc, R, user_locs=None):
    i_vec = np.array((1,0,0))
    j_vec = np.array((0,1,0))
    k_vec = np.array((0,0,1))

    et_vec2 = ( i_vec + np.sqrt(3.)*j_vec)/2.
    et_vec3 = (-i_vec + np.sqrt(3.)*j_vec)/2.

    alpha, beta, gamma, shape = det_angles_shape(loc, user_locs=user_locs)
    beta = np.pi/2 - beta
    EulerD1 = np.matmul(np.matmul(rot_mat(alpha,'k'), rot_mat(beta,'j')), rot_mat(gamma,'k'))

    if   shape == 'V3':
        eDArm1 = -1 * np.matmul(EulerD1,et_vec2)
        eDArm2 = -1 * np.matmul(EulerD1,et_vec3)
    elif shape == 'V2':
        eDArm1 =      np.matmul(EulerD1,et_vec3)
        eDArm2 = -1 * np.matmul(EulerD1,i_vec)
    elif shape == 'V1':
        eDArm1 =      np.matmul(EulerD1,i_vec)
        eDArm2 =      np.matmul(EulerD1,et_vec2)
    elif shape == 'L':
        eDArm1 =      np.matmul(EulerD1,i_vec)
        eDArm2 =      np.matmul(EulerD1,j_vec)

    return np.outer(eDArm1,eDArm1) - np.outer(eDArm2,eDArm2), R/cLight * np.matmul(EulerD1,k_vec)

def rot_mat(angle, axis):
    c = np.cos(angle)
    s = np.sin(angle)

    if axis == 'i': return np.array( ( (1,0,0), (0,c,-s), (0,s,c) ) )
    if axis == 'j': return np.array( ( (c,0,s), (0,1,0), (-s,0,c) ) )
    if axis == 'k': return np.array( ( (c,-s,0), (s,c,0), (0,0,1) ) )

def correct_arm_azimuth(arm_azimuth, which_arm):
    if   which_arm == 'x':        return np.pi/2 + arm_azimuth
    elif which_arm == 'y':        return           arm_azimuth
    elif which_arm == 'bisector': return np.pi/4 + arm_azimuth
    else: sys.exit('Unknown interferometer arm, specify which_arm as: "x", "y", or "bisect"!')

def det_angles_shape(loc, user_locs=None):
    # return alpha, beta, gamma in radians and shape of interferometer (e.g. 'L', 'V1', ...)
    # alpha ... longitude
    # beta  ... latitude
    # gamma ... angle from 'Due East' to y-arm
    # Switched to more precise angles for H, L, V, K, I, ET1, ET2, ET3 on 2022_03_29, found at:
    # https://lscsoft.docs.ligo.org/lalsuite/lal/_l_a_l_detectors_8h_source.html
    if user_locs is not None and loc in user_locs:
        user_loc = user_locs[loc]
        return user_loc['longitude'], user_loc['latitude'], correct_arm_azimuth(user_loc['arm_azimuth'], user_loc['which_arm']), user_loc['shape']
    # 2G sites
    elif loc == 'H':     return -2.08405676917,  0.81079526383, correct_arm_azimuth(np.pi-5.65487724844, 'y'), 'L'
    elif loc == 'L':     return -1.58430937078,  0.53342313506, correct_arm_azimuth(np.pi-4.40317772346, 'y'), 'L'
    elif loc == 'V':     return  0.18333805213,  0.76151183984, correct_arm_azimuth(np.pi-0.33916285222, 'y'), 'L'
    elif loc == 'K':     return  2.396441015,    0.6355068497,  correct_arm_azimuth(np.pi-1.054113,      'y'), 'L'
    elif loc == 'I':     return  1.33401332494,  0.24841853020, correct_arm_azimuth(np.pi-1.57079637051, 'y'), 'L'
    elif loc == 'LHO':   return -2.06982474503,  0.81079270401, correct_arm_azimuth(2.19910516123923,    'x'), 'L'
    elif loc == 'LLO':   return -1.5572845695,   0.53342110078, correct_arm_azimuth(3.45080197126464,    'x'), 'L'
    elif loc == 'LIO':   return  1.3444416672,   0.34231239582, correct_arm_azimuth(2.05248780779809,    'x'), 'L'
    # fiducial ET sites
    elif loc == 'ET1':   return  0.18333805213,  0.76151183984, correct_arm_azimuth(np.pi-0.33916285222, 'y'), 'V1'
    elif loc == 'ET2':   return  0.18333805213,  0.76151183984, correct_arm_azimuth(np.pi-0.33916285222, 'y'), 'V2'
    elif loc == 'ET3':   return  0.18333805213,  0.76151183984, correct_arm_azimuth(np.pi-0.33916285222, 'y'), 'V3'
    elif loc == 'ETS1':  return  0.1643518379,   0.70714923527, correct_arm_azimuth(1.5707963267949,     'x'), 'V1'
    elif loc == 'ETS2':  return  0.1643518379,   0.70714923527, correct_arm_azimuth(1.5707963267949,     'x'), 'V2'
    elif loc == 'ETS3':  return  0.1643518379,   0.70714923527, correct_arm_azimuth(1.5707963267949,     'x'), 'V3'
    # fiducial CE sites
    elif loc == 'C':     return -1.9691740,      0.764918,      correct_arm_azimuth(0.,                  'y'), 'L'
    elif loc == 'N':     return -1.8584265,      0.578751,      correct_arm_azimuth(-np.pi/3.,           'y'), 'L'
    elif loc == 'S':     return  2.5307270,     -0.593412,      correct_arm_azimuth(np.pi/4.,            'y'), 'L'
    elif loc == 'CEA':   return -2.18166156499,  0.80285145592, correct_arm_azimuth(4.53785605518526,    'x'), 'L'
    elif loc == 'CEB':   return -1.64060949687,  0.50614548308, correct_arm_azimuth(3.49065850398866,    'x'), 'L'
    elif loc == 'CES':   return  2.5307270,     -0.593412,      correct_arm_azimuth(np.pi/4.,            'y'), 'L'
    else: log_msg(f'det_angles_shape: Provided location {loc} is neither among the implemented locations nor in provided user_locs.', level='ERROR')



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import numpy as np
from gwbench import network

############################################################################
### User Choices
############################################################################

# choose the desired detectors
network_spec = ['aLIGO_H','aLIGO_L','aLIGO_V']
# initialize the network with the desired detectors
net = network.Network(network_spec)

# choose the desired waveform
wf_model_name = 'tf2'
# pass the chosen waveform to the network for initialization
net.set_wf_vars(wf_model_name=wf_model_name)

# pick the desired frequency range
f = np.arange(5.,61.5,2**-4)

# set the injection parameters
inj_params = {
    'Mc':    30.9,
    'eta':   0.247,
    'chi1z': 0,
    'chi2z': 0,
    'DL':    475,
    'tc':    0,
    'phic':  0,
    'iota':  np.pi/4,
    'ra':    np.pi/4,
    'dec':   np.pi/4,
    'psi':   np.pi/4,
    'gmst0': 0
    }

# assign with respect to which parameters to take derivatives
deriv_symbs_string = 'Mc eta DL tc phic iota ra dec psi'

# assign which parameters to convert to cos or log versions
conv_cos = ('iota','dec')
conv_log = ('Mc','DL')

# choose whether to take Earth's rotation into account
use_rot = 0

# pass all these variables to the network
net.set_net_vars(
    f=f, inj_params=inj_params,
    deriv_symbs_string=deriv_symbs_string,
    conv_cos=conv_cos, conv_log=conv_log,
    use_rot=use_rot
    )

############################################################################
### GW benchmarking
############################################################################

# compute the WF polarizations
net.calc_wf_polarizations()
# compute the WF polarizations and their derivatives
net.calc_wf_polarizations_derivs_num()

# setup antenna patterns, location phase factors, and PSDs
net.setup_ant_pat_lpf_psds()

# compute the detector responses
net.calc_det_responses()
# compute the detector responses and their derivatives
net.calc_det_responses_derivs_num()

# calculate the network and detector SNRs
net.calc_snrs()

# calculate the network and detector Fisher matrices, condition numbers,
# covariance matrices, error estimates, and inversion errors
net.calc_errors()

# calculate the 90%-credible sky area (in deg)
net.calc_sky_area_90()

############################################################################
### Print results
############################################################################

# print the contents of the detector objects (inside the network)
net.print_detectors()

# print the contents of the network objects
net.print_network()



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


"""This module handles calculations for a single gravitational wave detector.

"""

from copy import copy

import dill
import numpy as np

import gwbench.antenna_pattern_np as ant_pat_np
import gwbench.detector_response_derivatives as drd
import gwbench.err_deriv_handling as edh
import gwbench.fisher_analysis_tools as fat
import gwbench.psd as psd
import gwbench.snr as snr_mod
import gwbench.utils as utils

class Detector:

    ###
    #-----Init methods-----
    def __init__(self, det_key):
        #-----detector specification-----
        # full detector specification, specifying technology and location, e.g. CE2-40-CBO_C
        self.det_key = det_key
        # detector technology and locations
        self.tec = det_key.split('_')[0]
        self.loc = det_key.split('_')[1]

        #-----waveform and injection based quantities-----
        # frequency array
        self.f = None

        #-----technology and location based quantities-----
        # detector PSD
        self.psd = None
        # antenna pattern
        self.Fp = None
        self.Fc = None
        # location phase factor
        self.Flp = None

        #-----detector reponses-----
        # detector repsonse
        self.hf = None
        # derivative dictionary for detector responses
        self.del_hf = None
        # sympy expression of derivative dictionary for detector responses
        self.del_hf_expr = None

        #-----SNR-----
        # SNR, SNR^2 and d(SNR^2) calculated from self.hf
        self.snr = None
        self.snr_sq = None
        self.d_snr_sq = None

        #-----errors-----
        # Fisher matrix
        self.fisher = None
        # condition number of Fisher matrix
        self.cond_num = None
        # covariance matrix
        self.cov = None
        # dictionary containing information about the inversion error between the two matrices
        self.inv_err = None
        # dictionary of errors for given derivative variables
        self.errs = None


    ###
    #-----Setter methods-----
    def set_f(self, f):
        self.f = copy(f)


    ###
    #-----PSDs and antenna patterns-----
    def setup_psds(self, F_lo=-np.inf, F_hi=np.inf, user_psds=None):
        if user_psds is None or (self.det_key not in user_psds and self.tec not in user_psds):
            psd_file = None
            is_asd   = None
        elif self.det_key in user_psds:
            psd_file = user_psds[self.det_key]['psd_file']
            is_asd   = user_psds[self.det_key]['is_asd']
        else:
            psd_file = user_psds[self.tec]['psd_file']
            is_asd   = user_psds[self.tec]['is_asd']
        self.psd, self.f = psd.psd(self.tec,self.f,F_lo,F_hi,psd_file,is_asd)

    def setup_ant_pat_lpf(self, inj_params, use_rot, user_locs=None):
        self.Fp, self.Fc, self.Flp = ant_pat_np.antenna_pattern_and_loc_phase_fac(self.f, inj_params.get('Mc'), inj_params.get('tc'),
            inj_params['ra'], inj_params['dec'], inj_params['psi'], inj_params['gmst0'], self.loc, use_rot, user_locs=user_locs)


    ###
    #-----Detector responses-----
    def calc_det_responses(self, wf, inj_params):
        hfp, hfc = wf.eval_np_func(self.f,utils.get_sub_dict(inj_params,wf.wf_symbs_string))
        self.hf = self.Flp * (hfp * self.Fp + hfc * self.Fc)
        return self.hf, hfp, hfc

    def calc_det_responses_derivs_num(self, inj_params, deriv_variables, wf, deriv_symbs_string, conv_cos, conv_log, use_rot,
                                      step, method, order, d_order_n, user_locs, ana_deriv_symbs_string):
        _, hfp, hfc = self.calc_det_responses(wf, inj_params)
        self.del_hf = drd.calc_det_responses_derivs_num(self.loc, wf, deriv_symbs_string, self.f, inj_params, use_rot=use_rot, label='hf',
                                                        step=step, method=method, order=order, d_order_n=d_order_n, user_locs=user_locs,
                                                        ana_deriv_symbs_string=ana_deriv_symbs_string,
                                                        ana_deriv_aux={'hf':self.hf,   'hfp':hfp,         'hfc':hfc,
                                                                       'Fp':self.Fp,   'Fc':self.Fc,      'Flp':self.Flp,
                                                                       'loc':self.loc, 'use_rot':use_rot, 'user_locs':user_locs})
        self.del_hf, c_quants = edh.get_conv_del_eval_dic(self.del_hf, inj_params, conv_cos, conv_log, deriv_symbs_string)
        inj_params, deriv_variables = edh.get_conv_inj_params_deriv_variables(c_quants, inj_params, deriv_variables)

    def load_det_responses_derivs_sym(self, wf_model_name, deriv_symbs_string, use_rot, return_bin=0, user_lambdified_functions_path=None):
        self.del_hf_expr = drd.load_det_responses_derivs_sym(self.loc, wf_model_name, deriv_symbs_string, use_rot, return_bin, user_lambdified_functions_path)

    def calc_det_responses_derivs_sym(self, wf, inj_params, deriv_variables, conv_cos, conv_log, deriv_symbs_string):
        self.calc_det_responses(wf, inj_params)
        self.del_hf = {}
        for deriv in self.del_hf_expr:
            if deriv in ('variables', 'deriv_variables'): continue
            self.del_hf[deriv] = self.del_hf_expr[deriv](self.f, **utils.get_sub_dict(inj_params, self.del_hf_expr['variables']))

        self.del_hf, c_quants = edh.get_conv_del_eval_dic(self.del_hf, inj_params, conv_cos, conv_log, deriv_symbs_string)
        inj_params, deriv_variables = edh.get_conv_inj_params_deriv_variables(c_quants, inj_params, deriv_variables)


    ###
    #-----SNR calculations-----
    def calc_snrs(self, only_net):
        snr,snr_sq = snr_mod.snr_snr_sq_freq_array(self.hf, self.psd, self.f)
        if not only_net:
            self.snr = snr
            self.snr_sq = snr_sq
        return snr_sq

    def calc_snr_sq_integrand(self):
        self.d_snr_sq = snr_mod.snr_square_integrand(self.hf, self.psd)


    ###
    #-----Error calculation and Fisher analysis-----
    def calc_fisher_cov_matrices(self, only_net, cond_sup, logger=None):
        del_hf_sub_dict = utils.get_sub_dict(self.del_hf,('hf',),0)
        if not only_net:
            self.fisher, self.cov, self.cond_num, self.inv_err = fat.calc_fisher_cov_matrices(list(del_hf_sub_dict.values()), self.psd, self.f, only_fisher=0, cond_sup=cond_sup, logger=logger, tag=self.det_key)
            return self.fisher
        else:
            fisher,_,_,_ = fat.calc_fisher_cov_matrices(list(del_hf_sub_dict.values()), self.psd, self.f, only_fisher=1, cond_sup=cond_sup, logger=logger, tag=self.det_key)
            return fisher

    def calc_inv_err(self):
        self.inv_err = fat.inv_err_from_fisher_cov(self.fisher,self.cov)

    def calc_errs(self, deriv_variables):
        self.errs = fat.get_errs_from_cov(self.cov,deriv_variables)

    def calc_sky_area_90(self, deriv_variables, logger=None):
        if self.cov is None or self.errs is None: return
        if 'ra' in deriv_variables and ('cos_dec' in deriv_variables or 'dec' in deriv_variables):
            if 'cos_dec' in deriv_variables: dec_str = 'cos_dec'
            else:                                 dec_str = 'dec'
            ra_id      = deriv_variables.index('ra')
            dec_id     = deriv_variables.index(dec_str)
            is_cos_dec = (dec_str == 'cos_dec')
            self.errs['sky_area_90'] = edh.sky_area_90(self.errs['ra'],self.errs[dec_str],self.cov[ra_id,dec_id],self.inj_params['dec'],is_cos_dec)
        else:
            utils.log_msg(f'calc_sky_area_90: tag = {self.det_key} - Nothing done due to missing of either RA or COS_DEC (DEC) errors.', logger=logger, level='WARNING')

    def calc_sky_area_90_network(self, ra_id, dec_id, dec_val, is_cos_dec, dec_str):
        if self.cov is None or self.errs is None: return
        self.errs['sky_area_90'] = edh.sky_area_90(self.errs['ra'],self.errs[dec_str],self.cov[ra_id,dec_id],dec_val,is_cos_dec)


    ###
    #-----IO methods-----
    def print_detector(self,print_format=1):
        if print_format:
            sepl='-----------------------------------------------------------------------------------'
            print()
            print(sepl)
            print('Printing detector.')
            print(sepl)
            print()
        for key,value in vars(self).items():
            if type(value) == dict:
                print('Key: ',key)
                for key in value.keys():
                    print('',key)
                    print('',value[key])
                print()
            elif value is not None:
                print('Key: ',key)
                print(value)
                print()
        if print_format:
            print(sepl)
            print('Printing detector done.')
            print(sepl)
            print()



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


'''This module contains two methods that calculate the lambdified functions
of the the derivatives as well as of the function/expression itself.

 Input:  sympy function/expression, string of variables, boolean that determines
         if the function/expression has one (hf) or two (hfp, hfc) outputs.
 Output: dictionary that contains the lambdified functions:
         output['hf'] / output['hfp'], output['hfc']
         output['del_x_hf'] / output['del_x_hfp'], output['del_x_hfc']
           (x being the variable wrt which the derivative was taken)

 Disclaimer: These methods can be used on any sympy functions/expressions
             that have one or two outputs only (labels are set with
             gravitational waveforms in mind.
'''

from sympy import symbols, lambdify, diff

# hf is a sympy expression
def part_deriv_hf_expr(hf, symbols_string, deriv_symbs_string=None, pl_cr=0, label='hf'):
    symb_dic = {}

    for name in symbols_string.split(' '):
        symb_dic[name] = symbols(name,real=True)

    symb_list = list(symb_dic.values())

    if deriv_symbs_string == None:
        deriv_symbs_list = list(symb_dic.keys())
    else:
        deriv_symbs_list = deriv_symbs_string.split(' ')

    if pl_cr:
        lamdified_dic = {}
        for name in deriv_symbs_list:
            if name == 'f': continue

            key_string = 'del_'+name+'_'+label+'p'
            lamdified_dic[key_string] = lambdify(symb_list, diff(hf[0],symb_dic[name]), modules='numpy')

            key_string = 'del_'+name+'_'+label+'c'
            lamdified_dic[key_string] = lambdify(symb_list, diff(hf[1],symb_dic[name]), modules='numpy')

    else:
        lamdified_dic = {}
        for name in deriv_symbs_list:
            if name == 'f': continue

            key_string = 'del_'+name+'_'+label
            lamdified_dic[key_string] = lambdify(symb_list, diff(hf,symb_dic[name]), modules='numpy')

    return lamdified_dic



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


"""This module handles the benchmarking of graviational waveforms observed by a network of detectors.

"""

import logging
from copy import copy, deepcopy

import dill
import numpy as np

import gwbench.network as nc
import gwbench.utils as utils

# logger
glob_logger = logging.getLogger('multi_network_module')
glob_logger.setLevel('INFO')

class MultiNetwork:

    ###
    #-----Init methods-----
    def __init__(self, network_specs=None, logger_name='MultiNetwork', logger_level='WARNING', logger_level_network='WARNING'):
        ##-----logger-----
        self.logger = utils.get_logger(name=logger_name, level=logger_level)

        ##-----initialize network object-----
        if network_specs is None:
            # network_specs list and list of networks
            self.network_specs = None
            self.networks = None
            # dummy networks for unique locations
            self.loc_net  = None
        elif isinstance(network_specs, list):
            self.set_multi_network_from_specs(network_specs, logger_level_network=logger_level_network)

        if self.networks is None: self.logger.debug('Empty MultiNetwork initialized.')
        else:                     self.logger.debug('MultiNetwork initialized.')

    ###
    #-----Setter methods-----
    #
    # it is best practice to always change the instance variables using these setter methods
    #
    def set_logger_level(self, level):
        utils.set_logger_level(self.logger, level)

    def set_multi_network_from_specs(self, network_specs, logger_level_network='WARNING'):
        # copy network_specs
        self.network_specs = copy(network_specs)
        # prepare list of all networks specified in network_specs
        self.networks = [ nc.Network(network_spec, logger_name='..'.join(network_spec), logger_level=logger_level_network)
                          for network_spec in network_specs ]
        # find unique locations and initialize Network for these with dummy technologies
        self.loc_net = nc.Network(
            list(dict.fromkeys([ f'tec_{det_key.split("_")[1]}' for network_spec in network_specs for det_key in network_spec ])),
            logger_name='unique_locs', logger_level=logger_level_network)

    def set_wf_vars(self, wf_model_name, wf_other_var_dic=None, user_waveform=None, cosmo=None):
        for net in [*self.networks, self.loc_net]:
            net.set_wf_vars(wf_model_name, wf_other_var_dic=wf_other_var_dic, user_waveform=user_waveform, cosmo=cosmo)

    def set_net_vars(self, f=None, inj_params=None, deriv_symbs_string=None, conv_cos=None, conv_log=None,
                     use_rot=None, user_locs=None, user_psds=None, user_lambdified_functions_path=None, ana_deriv_symbs_string=None):
        for net in [*self.networks, self.loc_net]:
            net.set_net_vars(f=f, inj_params=inj_params, deriv_symbs_string=deriv_symbs_string, conv_cos=conv_cos, conv_log=conv_log,
                             use_rot=use_rot, user_locs=user_locs, user_psds=user_psds, ana_deriv_symbs_string=ana_deriv_symbs_string,
                             user_lambdified_functions_path=user_lambdified_functions_path)


    ###
    #-----Resetter methods for instance variables-----
    def reset_ant_pat_lpf_psds(self):
        for net in [*self.networks, self.loc_net]:
            net.reset_ant_pat_lpf_psds()

    def reset_wf_polarizations(self):
        for net in [*self.networks, self.loc_net]:
            net.reset_wf_polarizations()

    def reset_det_responses(self):
        for net in [*self.networks, self.loc_net]:
            net.reset_det_responses()

    def reset_snrs(self):
        for net in [*self.networks, self.loc_net]:
            net.reset_snrs()

    def reset_errors(self):
        for net in [*self.networks, self.loc_net]:
            net.reset_errors()

    ###
    #-----PSDs and antenna patterns-----
    def setup_ant_pat_lpf_psds(self, F_lo=-np.inf, F_hi=np.inf):
        self.setup_psds(F_lo, F_hi)
        self.setup_ant_pat_lpf()

    def setup_psds(self, F_lo=-np.inf, F_hi=np.inf):
        for net in [*self.networks, self.loc_net]:
            net.setup_psds(F_lo=F_lo, F_hi=F_hi)
        self.logger.info('PSDs loaded.')

    def setup_ant_pat_lpf(self):
        for net in [*self.networks, self.loc_net]:
            net.setup_ant_pat_lpf()
        self.logger.info('Antenna patterns and LPFs loaded.')


    ###
    #-----Waveform polarizations-----
    def calc_wf_polarizations(self):
        self.loc_net.calc_wf_polarizations()
        for net in self.networks:
            net.hfp = self.loc_net.hfp
            net.hfc = self.loc_net.hfc

    def calc_wf_polarizations_derivs_num(self, step=1e-9, method='central', order=2, d_order_n=1):
        self.loc_net.calc_wf_polarizations_derivs_num(step=step, method=method, order=order, d_order_n=d_order_n)

    def load_wf_polarizations_derivs_sym(self, return_bin=0):
        self.loc_net.load_wf_polarizations_derivs_sym(return_bin=return_bin)

    def calc_wf_polarizations_derivs_sym(self):
        self.loc_net.calc_wf_polarizations_derivs_sym()

    def dist_wf_polarizations(self):
        for net in self.networks:
            net.hfp             =     copy(self.loc_net.hfp)
            net.hfc             =     copy(self.loc_net.hfc)
            net.del_hfpc        = deepcopy(self.loc_net.del_hfpc)
            net.del_hfpc_expr   = deepcopy(self.loc_net.del_hfpc_expr)
            net.inj_params      = deepcopy(self.loc_net.inj_params)
            net.deriv_variables =     copy(self.loc_net.deriv_variables)
        self.logger.info('Polarizations distributed among all networks.')


    ###
    #-----Detector responses-----
    def calc_det_responses(self):
        self.loc_net.calc_det_responses()

    def calc_det_responses_derivs_num(self, step=1e-9, method='central', order=2, d_order_n=1, num_cores=None):
        self.loc_net.calc_det_responses_derivs_num(step=step, method=method, order=order, d_order_n=d_order_n, num_cores=num_cores)

    def load_det_responses_derivs_sym(self, return_bin=0):
        self.loc_net.load_det_responses_derivs_sym(return_bin=return_bin)

    def calc_det_responses_derivs_sym(self, num_cores=None):
        self.loc_net.calc_det_responses_derivs_sym(num_cores=num_cores)

    def dist_det_responses(self):
        for net in self.networks:
            net.inj_params          = deepcopy(self.loc_net.inj_params)
            net.deriv_variables     =     copy(self.loc_net.deriv_variables)
            for det in net.detectors:
                loc_det             = self.loc_net.get_detector(f'tec_{det.loc}') 
                f_ids               = np.logical_and(loc_det.f >= det.f[0], loc_det.f <= det.f[-1])
                det.hf              = copy(loc_det.hf[f_ids])
                det.del_hf          = { deriv : copy(del_hf_deriv[f_ids]) for deriv,del_hf_deriv in loc_det.del_hf.items() }
                det.del_hf_expr     = deepcopy(loc_det.del_hf_expr)
        self.logger.info('Detector responses distributed among all networks.')


    ###
    #-----SNR calculations-----
    def calc_snrs(self, only_net=0):
        for net in self.networks:
            net.calc_snrs(only_net=only_net)
        self.logger.info('SNRs calculated.')

    def calc_snr_sq_integrand(self):
        for net in self.networks:
            net.calc_snr_sq_integrand()
        self.logger.info('SNR integrands calculated.')


    ###
    #-----Error calculation and Fisher analysis-----
    def calc_errors(self, cond_sup=None, only_net=0):
        self.logger.info('Calculate errors (Fisher & cov matrices).')
        for net in self.networks:
            net.calc_errors(cond_sup=cond_sup, only_net=only_net)
        self.logger.info('Errors calculated.')

    def calc_sky_area_90(self, only_net=0):
        for net in self.networks:
            net.calc_sky_area_90(only_net=only_net)
        self.logger.info('Sky areas calculated.')


    ###
    #-----IO methods-----
    def save_multi_network(self, filename_path):
        '''Save the network under the given path using *dill*.'''
        with open(filename_path, "wb") as fi:
            dill.dump(self, fi, recurse=True)
        self.logger.info('MultiNetwork pickled.')
        return

    def load_multi_network(filename_path):
        '''Loading the network from the given path using *dill*.'''
        with open(filename_path, "rb") as fi:
            mul_net = dill.load(fi)
        mul_net.logger.info('MultiNetwork loaded.')
        return mul_net



from copy import deepcopy

import dill
import numpy as np

import gwbench.detector as dc
import gwbench.detector_response_derivatives as drd
import gwbench.err_deriv_handling as edh
import gwbench.network as nc
import gwbench.utils as utils

###
#-----Dealing with several networks-----
def get_det_responses_psds_from_locs_tecs(net,loc_net,tec_net,F_lo=-np.inf,F_hi=np.inf,sym_derivs=0,keep_variables=None):

    net.inj_params = loc_net.inj_params
    net.deriv_variables = loc_net.deriv_variables
    net.f = loc_net.f

    for i,det in enumerate(net.detectors):
        tec_det = tec_net.get_detector(det.tec+'_loc')

        f_lo = np.maximum(tec_det.f[0], F_lo)
        f_hi = np.minimum(tec_det.f[-1], F_hi)
        ids_det_f = np.logical_and(tec_det.f>=f_lo,tec_det.f<=f_hi)
        ids_net_f = np.logical_and(net.f>=f_lo,net.f<=f_hi)

        det.f = tec_det.f[ids_det_f]
        det.psd = tec_det.psd[ids_det_f]

        loc_det = loc_net.get_detector('tec_'+det.loc)
        det.hf = deepcopy(loc_det.hf[ids_net_f])
        det.del_hf = deepcopy(loc_det.del_hf)
        for deriv in det.del_hf:
            det.del_hf[deriv] = det.del_hf[deriv][ids_net_f]
        if sym_derivs:
            det.del_hf_expr = deepcopy(loc_det.del_hf_expr)

    # only keep derivs defined by the the variables in keep_variables
    if keep_variables is not None:
        keep_derivs = []
        keep_deriv_variables = []

        for keep_variable in keep_variables:
            k = 0
            for i,deriv in enumerate(list(net.detectors[0].del_hf.keys())):
                if keep_variable in deriv:
                    keep_derivs.append(deriv)
                    keep_deriv_variables.append(net.deriv_variables[i])
                    k = 1
            if not k:
                net.logger.warning(f'get_det_responses_psds_from_locs_tecs: {keep_variable} not among the derivatives.')

        net.deriv_variables = keep_deriv_variables
        for det in net.detectors:
            det.del_hf = utils.get_sub_dict(det.del_hf,keep_derivs,keep_in_list=1)

    net.logger.info('Detector responses transferred.')


###
#-----Dealing with several networks-----
def unique_tecs(network_specs, f, F_lo=-np.inf, F_hi=np.inf, user_psds=None, logger_level='WARNING'):
    # initialize empty network
    tec_net = nc.Network(logger_name='unique_tecs', logger_level=logger_level)
    # get the detector keys
    tec_net.det_keys = []

    tec_net.logger.info('Calculate PSDs for unique detector technologies.')

    # find unique technologies
    for network_spec in network_specs:
        network = nc.Network(network_spec, logger_name='unique_tecs', logger_level=logger_level)
        for det in network.detectors:
            tec_net.det_keys.append(det.tec)
    tec_net.det_keys = list(dict.fromkeys(tec_net.det_keys))

    # make them into fake detector keys
    for i,tec in enumerate(tec_net.det_keys):
        tec_net.det_keys[i] = tec+'_loc'
    # initialize fake detectors
    tec_net.detectors = []
    for det_key in tec_net.det_keys:
        tec_net.detectors.append(dc.Detector(det_key))

    # get PSDs
    tec_net.set_net_vars(f=f, user_psds=user_psds)
    tec_net.setup_psds(F_lo, F_hi)

    tec_net.logger.info('PSDs for unique detector technologies calculated.')
    return tec_net


def unique_locs_det_responses(network_specs, f, inj_params, deriv_symbs_string, wf_model_name,
                              wf_other_var_dic=None, conv_cos=None, conv_log=None, use_rot=1,
                              user_waveform=None, user_locs=None, user_lambdified_functions_path=None,
                              ana_deriv_symbs_string=None, cosmo=None, logger_level='WARNING', num_cores=None,
                              step=None, method=None, order=None, d_order_n=None):
    # initialize empty network
    loc_net = nc.Network(logger_name='unique_locs_det_responses', logger_level=logger_level)
    # get the detector keys
    loc_net.det_keys = []

    # find unique locations
    for network_spec in network_specs:
        network = nc.Network(network_spec, logger_name='unique_locs_det_responses', logger_level=logger_level)
        for det in network.detectors:
            loc_net.det_keys.append(det.loc)
    loc_net.det_keys = list(dict.fromkeys(loc_net.det_keys))

    # make them into fake detectpr keys
    for i,loc in enumerate(loc_net.det_keys):
        loc_net.det_keys[i] = 'tec_'+loc
    # initialize fake detectors
    loc_net.detectors = []
    for det_key in loc_net.det_keys:
        loc_net.detectors.append(dc.Detector(det_key))
    # set all the other necessary variables
    loc_net.set_wf_vars(wf_model_name, wf_other_var_dic=wf_other_var_dic, user_waveform=user_waveform, cosmo=cosmo)
    loc_net.set_net_vars(f=f, inj_params=inj_params, deriv_symbs_string=deriv_symbs_string, conv_cos=conv_cos, conv_log=conv_log,
                         use_rot=use_rot, user_locs=user_locs, user_lambdified_functions_path=user_lambdified_functions_path)

    # setup Fp, Fc, and Flp and calculate the detector responses
    loc_net.setup_ant_pat_lpf()
    loc_net.calc_det_responses()

    if step is None:
        loc_net.logger.info('Loading the lamdified functions.')
        loc_net.load_det_responses_derivs_sym(return_bin = 1)
        loc_net.logger.info('Loading done.')

    loc_net.logger.info('Evaluate lambdified detector responses for unique locations.')
    if num_cores is None:
        if step is None:
            for det in loc_net.detectors:
                det.del_hf, c_quants = eval_loc_sym(det.loc,det.del_hf_expr,deriv_symbs_string,f,inj_params,conv_cos,conv_log,logger=loc_net.logger)
        else:
            for det in loc_net.detectors:
                det.del_hf, c_quants = eval_loc_num(det.loc,loc_net.wf,deriv_symbs_string,f,inj_params,conv_cos,conv_log,use_rot,
                                                    step,method,order,d_order_n,user_locs,ana_deriv_symbs_string,logger=loc_net.logger)
        loc_net.inj_params, loc_net.deriv_variables = edh.get_conv_inj_params_deriv_variables(c_quants, loc_net.inj_params, loc_net.deriv_variables)

    else:
        from multiprocessing import Pool
        pool = Pool(num_cores)
        if step is None:
            arg_tuple_list = [(det.loc,det.del_hf_expr,deriv_symbs_string,f,inj_params,conv_cos,conv_log,loc_net.logger) for det in loc_net.detectors]
            result = pool.starmap_async(eval_loc_sym, arg_tuple_list)
            result.wait()
        else:
            arg_tuple_list = [(det.loc,loc_net.wf,deriv_symbs_string,f,inj_params,conv_cos,conv_log,use_rot,
                               step,method,order,d_order_n,user_locs,ana_deriv_symbs_string,loc_net.logger) for det in loc_net.detectors]
            result = pool.starmap_async(eval_loc_num, arg_tuple_list)
            result.wait()

        for det, (del_hf,c_quants) in zip(loc_net.detectors, result.get()):
            det.del_hf = del_hf

        loc_net.inj_params, loc_net.deriv_variables = edh.get_conv_inj_params_deriv_variables(c_quants, loc_net.inj_params, loc_net.deriv_variables)

    if step is None:
        for det in loc_net.detectors:
            det.del_hf_expr = dill.loads(det.del_hf_expr)

    loc_net.logger.info('Lambdified detector responses for unique locations evaluated.')
    return loc_net


def eval_loc_sym(loc ,del_hf_expr, deriv_symbs_string, f, inj_params, conv_cos, conv_log, logger=None):
    if logger is None: glob_logger.info(f'   {loc}')
    else:              logger.info(f'   {loc}')
    del_hf = {}
    del_hf_expr = dill.loads(del_hf_expr)
    for deriv in del_hf_expr:
        if deriv in ('variables','deriv_variables'): continue
        del_hf[deriv] = del_hf_expr[deriv](f,**utils.get_sub_dict(inj_params,del_hf_expr['variables']))
    return edh.get_conv_del_eval_dic(del_hf,inj_params,conv_cos,conv_log, deriv_symbs_string)


def eval_loc_num(loc, wf, deriv_symbs_string, f, inj_params, conv_cos, conv_log, use_rot,
                 step, method, order, d_order_n, user_locs, ana_deriv_symbs_string, logger=None):
    if logger is None: glob_logger.info(f'   {loc}')
    else:              logger.info(f'   {loc}')
    del_hf = drd.calc_det_responses_derivs_num(loc, wf, deriv_symbs_string, f, inj_params, use_rot=use_rot, label='hf',
                                               step=step, method=method, order=order, d_order_n=d_order_n, user_locs=user_locs,
                                               ana_deriv_symbs_string=ana_deriv_symbs_string)
    return edh.get_conv_del_eval_dic(del_hf,inj_params,conv_cos,conv_log, deriv_symbs_string)



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import numpy as np
import sympy as sp

from gwbench.antenna_pattern_np import ap_symbs_string, det_angles_shape
from gwbench.utils import time_fac, REarth, halfTEarth, cLight

cos = sp.cos
sin = sp.sin
exp = sp.exp

f, Mc, tc, ra, dec, psi, gmst0 = sp.symbols(ap_symbs_string, real=True)

def detector_response_expr(hfp, hfc, loc, use_rot, user_locs=None):
    # input:    hfp     sympy expression of plus polarization
    #           hfc     sympy expression of cross polarization
    #           loc     location (and implied orientation) of a detector
    #           use_rot use frequency dependent time due to roa=tation of earth and SPA
    #
    # output:   hf
    return detector_response(f, hfp, hfc, Mc, tc, ra, dec, psi, gmst0, loc, use_rot, user_locs=user_locs)

def antenna_pattern_and_loc_phase_fac_expr(loc, use_rot, user_locs=None):
    # input:    loc     location (and implied orientation) of a detector
    #           use_rot use frequency dependent time due to roa=tation of earth and SPA
    #
    # output:   Fp, Fc, Flp
    return antenna_pattern_and_loc_phase_fac(f, Mc, tc, ra, dec, psi, gmst0, loc, use_rot, user_locs=user_locs)

def detector_response(f, hfp, hfc, Mc, tc, ra, dec, psi, gmst0, loc, use_rot, user_locs=None):
    # input:    f       frequency domain [Hz]
    #           hfc     cross polarization
    #           hfp     plus polarization
    #           Mc      chirp Mass [solar mass]
    #           tc      time of coalescence [s]
    #           dec     declination [rad]
    #           ra      right ascencsion [rad]
    #           psi     polarization angle [rad]
    #           gmst0   GreenwichMeanSiderialTime according to LAL
    #           loc     location (and implied orientation) of a detector
    #           use_rot use frequency dependent time due to rotation of earth and SPA
    #
    # output:   hf      detector strain

    Fp, Fc, Flp = antenna_pattern_and_loc_phase_fac(f, Mc, tc, ra, dec, psi, gmst0, loc, use_rot, user_locs=user_locs)
    return Flp * (Fp * hfp + Fc * hfc)

def antenna_pattern_and_loc_phase_fac(f, Mc, tc, ra, dec, psi, gmst0, loc, use_rot, user_locs=None):
    # input:    f       frequency domain [Hz]
    #           Mc      chirp Mass [solar mass]
    #           tc      time of coalescence [s]
    #           dec     declination [rad]
    #           ra      right ascencsion [rad]
    #           psi     polarization angle [rad]
    #           gmst0   GreenwichMeanSiderialTime according to LAL
    #           loc     location (and implied orientation) of a detector
    #           use_rot use frequency dependent time due to rotation of earth and SPA
    #
    # output:   Fp, Fc, Flp

    gra  = calc_gra(ra, gmst0, use_rot, f, Mc, tc)
    D, d = det_ten_and_loc_vec(loc, REarth, user_locs=user_locs)

    return *ant_pat_funcs(D, *ant_pat_vectors(gra, dec, psi)), loc_phase_func(gra, dec, f, d)

def calc_gra(ra, gmst0, use_rot, f, Mc, tc):
    if use_rot: return gmst0 - ra + np.pi / halfTEarth * (tc - 5. / 256 * (time_fac * Mc)**(-5./3) * (np.pi * f)**(-8./3))
    else:       return gmst0 - ra

def loc_phase_func(gra, dec, f, d):
    theta = np.pi/2 - dec
    return exp(1j * 2 * np.pi * f * d.T*sp.Matrix([cos(gra)*sin(theta), sin(gra)*sin(theta), cos(theta)]))[0,0]

def ant_pat_funcs(D, XX, YY):
    return (0.5 * (XX.T*D*XX - YY.T*D*YY))[0,0], \
           (0.5 * (XX.T*D*YY + YY.T*D*XX))[0,0]

def ant_pat_vectors(gra, dec, psi):
    return sp.Matrix([ -cos(psi)*sin(gra) - sin(psi)*cos(gra)*sin(dec),
                       -cos(psi)*cos(gra) + sin(psi)*sin(gra)*sin(dec),
                                                     sin(psi)*cos(dec) ]), \
           sp.Matrix([  sin(psi)*sin(gra) - cos(psi)*cos(gra)*sin(dec),
                        sin(psi)*cos(gra) + cos(psi)*sin(gra)*sin(dec),
                                                     cos(psi)*cos(dec) ])

def det_ten_and_loc_vec(loc, R, user_locs=None):
    i_vec = sp.Matrix([1,0,0])
    j_vec = sp.Matrix([0,1,0])
    k_vec = sp.Matrix([0,0,1])

    et_vec2 = ( i_vec + np.sqrt(3.)*j_vec)/2.
    et_vec3 = (-i_vec + np.sqrt(3.)*j_vec)/2.

    alpha, beta, gamma, shape = det_angles_shape(loc, user_locs=user_locs)
    beta = np.pi/2 - beta
    EulerD1 = rot_mat(alpha,'k') * rot_mat(beta,'j') * rot_mat(gamma,'k')

    if   shape == 'V3':
        eDArm1 = -1 * EulerD1*et_vec2
        eDArm2 = -1 * EulerD1*et_vec3
    elif shape == 'V2':
        eDArm1 =      EulerD1*et_vec3
        eDArm2 = -1 * EulerD1*i_vec
    elif shape == 'V1':
        eDArm1 =      EulerD1*i_vec
        eDArm2 =      EulerD1*et_vec2
    elif shape == 'L':
        eDArm1 =      EulerD1*i_vec
        eDArm2 =      EulerD1*j_vec

    return eDArm1*eDArm1.T - eDArm2*eDArm2.T, R/cLight * EulerD1*k_vec

def rot_mat(angle, axis):
    c = np.cos(angle)
    s = np.sin(angle)

    if axis == 'i': return sp.Matrix( [ [1,0,0], [0,c,-s], [0,s,c] ] )
    if axis == 'j': return sp.Matrix( [ [c,0,s], [0,1,0], [-s,0,c] ] )
    if axis == 'k': return sp.Matrix( [ [c,-s,0], [s,c,0], [0,0,1] ] )



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import os
import warnings
from copy import copy

import astropy.cosmology as apcosm
import numpy as np
from scipy.integrate import quad, simps
from scipy.interpolate import interp1d
from scipy.optimize import minimize_scalar

import gwbench.basic_relations as brs

PI = np.pi

###
#-----standard values for reference-----
std_vals = [{'source':'BBH', 'dist':'power_uniform', 'mmin':5, 'mmax':100, 'alpha':-1.6},
            {'source':'BNS', 'dist':'gaussian', 'mean':1.35, 'sigma':0.15},
            {'source':'BNS', 'dist':'uniform', 'mmin':0.8, 'mmax':3}]

###
#-----CBC parameters sampler-----
def injections_CBC_params_redshift(cosmo_dict,mass_dict,spin_dict,redshifted,num_injs=10,seed=None,file_path=None):
    rng = np.random.default_rng(seed)
    seeds = rng.integers(100000,size=4)

    m1_vec, m2_vec = mass_sampler(mass_dict,num_injs,seeds[0])
    chi1x_vec, chi1y_vec, chi1z_vec, chi2x_vec, chi2y_vec, chi2z_vec = spin_sampler(spin_dict,num_injs,seeds[1])
    z_vec, DL_vec = redshift_lum_distance_sampler(cosmo_dict,num_injs,seeds[2])
    iota_vec, ra_vec, dec_vec, psi_vec = angle_sampler(num_injs,seeds[3])

    Mc_vec, eta_vec = get_Mc_eta(m1_vec,m2_vec)
    if redshifted: Mc_vec *= (1. + z_vec)

    params = [Mc_vec, eta_vec, chi1x_vec, chi1y_vec, chi1z_vec, chi2x_vec, chi2y_vec, chi2z_vec, DL_vec, iota_vec, ra_vec, dec_vec, psi_vec, z_vec]
    if file_path is not None: save_injections(params,file_path)
    return params


###
#-----IO functions-----
def load_injections(file_path):
    return np.transpose(np.loadtxt(file_path))

def save_injections(params,file_path):
    np.savetxt(os.path.join(file_path), np.transpose(np.array([params][0])), delimiter = ' ')


###
#-----angle samplers-----
def angle_sampler(num_injs,seed):
    rngs = [np.random.default_rng(seeed) for seeed in np.random.default_rng(seed).integers(100000,size=4)]
    iota_vec = np.arccos(rngs[0].uniform(low=-1, high=1, size=num_injs))
    ra_vec   = rngs[1].uniform(low=0., high=2.*PI, size=num_injs)
    dec_vec  = np.arccos(rngs[2].uniform(low=-1, high=1, size=num_injs)) - PI/2.
    psi_vec  = rngs[3].uniform(low=0., high=2.*PI, size=num_injs)
    return iota_vec, ra_vec, dec_vec, psi_vec

###
#-----spin samplers-----
def spin_sampler(spin_dict,num_injs,seed):
    if 'dist' not in spin_dict:
        rngs   = [np.random.default_rng(seeed) for seeed in np.random.default_rng(seed).integers(100000,size=6)]
        chi_lo = spin_dict['chi_lo']
        chi_hi = spin_dict['chi_hi']
        dim    = spin_dict['dim']

        if   dim == 1:
            chiz_vecs = [rngs[i].uniform(low=chi_lo, high=chi_hi, size=num_injs) for i in (2,5)]
            return [np.zeros(num_injs), np.zeros(num_injs), chiz_vecs[0], np.zeros(num_injs), np.zeros(num_injs), chiz_vecs[1]]
        elif dim == 3:
            if   spin_dict['geom'] == 'cartesian':
                return [rngs[i].uniform(low=chi_lo, high=chi_hi, size=num_injs) for i in range(6)]
            elif spin_dict['geom'] == 'spherical':
                # chi1
                chi_vec   = (rngs[0].uniform(low=chi_lo**3., high=chi_hi**3., size=num_injs))**(1./3.)
                theta_vec = np.arccos(rngs[1].uniform(low=-1, high=1, size=num_injs))
                phi_vec   = rngs[2].uniform(low=0., high=2.*PI, size=num_injs)
                chi1x_vec, chi1y_vec, chi1z_vec = get_cartesian_from_spherical(chi_vec,theta_vec,phi_vec)
                # chi2
                chi_vec   = (rngs[3].uniform(low=chi_lo**3., high=chi_hi**3., size=num_injs))**(1./3.)
                theta_vec = np.arccos(rngs[4].uniform(low=-1, high=1, size=num_injs))
                phi_vec   = rngs[5].uniform(low=0., high=2.*PI, size=num_injs)
                chi2x_vec, chi2y_vec, chi2z_vec = get_cartesian_from_spherical(chi_vec,theta_vec,phi_vec)
                return [chi1x_vec, chi1y_vec, chi1z_vec, chi2x_vec, chi2y_vec, chi2z_vec]

    elif spin_dict['dist'] == 'gaussian_uniform':
        rngs     = [np.random.default_rng(seeed) for seeed in np.random.default_rng(seed).integers(100000,size=2)]
        mean     = spin_dict['mean']
        sigma    = spin_dict['sigma']
        chi1_min = spin_dict['chi1_min']
        chi1_max = spin_dict['chi1_max']
        chi2_min = spin_dict['chi2_min']
        chi2_max = spin_dict['chi2_max']

        return [np.zeros(num_injs), np.zeros(num_injs), truncated_gaussian(mean, sigma, chi1_min, chi1_max, num_injs, rng=rngs[0]),
                np.zeros(num_injs), np.zeros(num_injs), rngs[1].uniform(low=chi2_min, high=chi2_max, size=num_injs)]

    elif spin_dict['dist'] == 'beta':
        rngs  = [np.random.default_rng(seeed) for seeed in np.random.default_rng(seed).integers(100000,size=2)]
        alpha = spin_dict['alpha']
        beta  = spin_dict['beta']

        return [np.zeros(num_injs), np.zeros(num_injs), rngs[0].beta(alpha, beta, num_injs),
                np.zeros(num_injs), np.zeros(num_injs), rngs[1].beta(alpha, beta, num_injs)]

    elif spin_dict['dist'] == 'beta_gaussian_uniform':
        rngs         = [np.random.default_rng(seeed) for seeed in np.random.default_rng(seed).integers(100000,size=9)]
        alpha        = spin_dict['alpha']
        beta         = spin_dict['beta']
        mean         = mass_dict['mean']
        sigma        = mass_dict['sigma']
        weight       = mass_dict['weight']
        cos_min      = mass_dict['cos_min']
        cos_max      = mass_dict['cos_max']

        chi1_vec     = rngs[0].beta(alpha, beta, num_injs)
        chi2_vec     = rngs[1].beta(alpha, beta, num_injs)

        azimuth1_vec = rngs[2].uniform(low=0., high=2.*PI, size=num_injs)
        azimuth2_vec = rngs[3].uniform(low=0., high=2.*PI, size=num_injs)

        N            = int(weight * num_injs)
        M            = num_injs - N
        tilt1_vec    = np.arccos(np.concatenate((truncated_gaussian(mean, sigma, cos_min, cos_max, N, rng=rngs[4]),
                                                 rngs[5].uniform(low=cos_min, high=cos_max, size=M))))
        tilt2_vec    = np.arccos(np.concatenate((truncated_gaussian(mean, sigma, cos_min, cos_max, N, rng=rngs[6]),
                                                 rngs[7].uniform(low=cos_min, high=cos_max, size=M))))
        ids          = np.arange(num_injs)
        rngs[8].shuffle(ids)
        tilt1_vec    = tilt1_vec[ids]
        tilt2_vec    = tilt2_vec[ids]

        chi1xy_vec   = chi1_vec * np.sin(tilt1_vec)
        chi2xy_vec   = chi2_vec * np.sin(tilt2_vec)

        return [chi1xy_vec * np.cos(azimuth1_vec), chi1xy_vec * np.sin(azimuth1_vec), chi1_vec * np.cos(tilt1_vec),
                chi2xy_vec * np.cos(azimuth2_vec), chi2xy_vec * np.sin(azimuth2_vec), chi2_vec * np.cos(tilt2_vec)]

def get_cartesian_from_spherical(r,theta,phi):
    return r * np.sin(theta) * np.cos(phi), r * np.sin(theta) * np.sin(phi), r * np.cos(theta)


###
#-----mass samplers-----
def mass_sampler(mass_dict,num_injs,seed):
    rngs = [np.random.default_rng(seeed) for seeed in np.random.default_rng(seed).integers(100000,size=5)]
    if mass_dict['dist'] == 'gaussian':
        mmin  = mass_dict['mmin']
        mmax  = mass_dict['mmax']
        mean  = mass_dict['mean']
        sigma = mass_dict['sigma']
        m1_m2 = 0

        m1_vec = truncated_gaussian(mean, sigma, mmin, mmax, num_injs, rng=rngs[0])
        m2_vec = truncated_gaussian(mean, sigma, mmin, mmax, num_injs, rng=rngs[1])

    elif mass_dict['dist'] == 'double_gaussian':
        mmin   = mass_dict['mmin']
        mmax   = mass_dict['mmax']
        mean1  = mass_dict['mean1']
        sigma1 = mass_dict['sigma1']
        mean2  = mass_dict['mean2']
        sigma2 = mass_dict['sigma2']
        weight = mass_dict['weight']
        m1_m2  = 0

        N = int(weight * num_injs)
        M = num_injs - N

        m1_vec = np.concatenate((truncated_gaussian(mean1, sigma1, mmin, mmax, N, rng=rngs[0]),
                                 truncated_gaussian(mean2, sigma2, mmin, mmax, M, rng=rngs[1])))

        m2_vec = np.concatenate((truncated_gaussian(mean1, sigma1, mmin, mmax, N, rng=rngs[2]),
                                 truncated_gaussian(mean2, sigma2, mmin, mmax, M, rng=rngs[3])))

        ids    = np.arange(num_injs)
        rngs[4].shuffle(ids)
        m1_vec = m1_vec[ids]
        m2_vec = m2_vec[ids]

    elif mass_dict['dist'] == 'lognormal':
        mass_scale = mass_dict['mass_scale']
        mean       = mass_dict['mean']
        sigma      = mass_dict['sigma']
        m1_m2      = 1

        m1_vec = mass_scale * rngs[0].lognormal(mean, sigma, num_injs)
        m2_vec = mass_scale * rngs[1].lognormal(mean, sigma, num_injs)

    elif mass_dict['dist'] == 'power':
        mmin  = mass_dict['mmin']
        mmax  = mass_dict['mmax']
        alpha = mass_dict['alpha'] + 1
        m1_m2 = 0

        m1_vec = (mmin**alpha + (mmax**alpha - mmin**alpha)*rngs[0].random(num_injs))**(1./alpha)
        m2_vec = (mmin**alpha + (mmax**alpha - mmin**alpha)*rngs[1].random(num_injs))**(1./alpha)

    elif mass_dict['dist'] == 'power_peak':
                                                # standard power+peak parameters from GTWC-2 populations paper:
                                                # https://arxiv.org/abs/2010.14533
        mmin       = mass_dict['mmin']          # mmin       = 4.59
        mmax       = mass_dict['mmax']          # mmax       = 86.22
        m1_alpha   = mass_dict['m1_alpha']      # m1_alpha   = 2.63
        peak_frac  = mass_dict['peak_frac']     # peak_frac  = 0.1
        peak_mean  = mass_dict['peak_mean']     # peak_mu    = 33.07
        peak_sigma = mass_dict['peak_sigma']    # peak_sigma = 5.69
        delta_m    = mass_dict['delta_m']       # delta_m    = 4.82
        q_beta     = mass_dict['q_beta']        # q_beta     = 1.26
        m1_m2      = 1

        m1_vec     = power_peak(mmin, mmax, m1_alpha, peak_frac, peak_mean, peak_sigma, delta_m, num_injs, nm1s=5001, rng=rngs[0])

        q_vec      = rngs[1].power(q_beta + 1, num_injs)
        m2_vec     = q_vec * m1_vec
        m2_mask    = m2_vec < mmin

        while m2_mask.sum():
            q_vec[m2_mask] = rngs[1].power(q_beta + 1, m2_mask.sum())
            m2_vec         = q_vec * m1_vec
            m2_mask        = m2_vec < mmin

    elif mass_dict['dist'] == 'power_peak_uniform':
        mmin       = mass_dict['mmin']
        mmax       = mass_dict['mmax']
        m1_alpha   = mass_dict['m1_alpha']
        peak_frac  = mass_dict['peak_frac']
        peak_mean  = mass_dict['peak_mean']
        peak_sigma = mass_dict['peak_sigma']
        delta_m    = mass_dict['delta_m']
        m1_m2      = 1

        m1_vec     = power_peak(mmin, mmax, m1_alpha, peak_frac, peak_mean, peak_sigma, delta_m, num_injs, nm1s=5001, rng=rngs[0])
        m2_vec     = rngs[1].uniform(mmin, m1_vec)

    elif mass_dict['dist'] == 'nsbh_power_peak_uniform':
        m1_min     = mass_dict['m1_min']
        m1_max     = mass_dict['m1_max']
        m1_alpha   = mass_dict['m1_alpha']
        peak_frac  = mass_dict['peak_frac']
        peak_mean  = mass_dict['peak_mean']
        peak_sigma = mass_dict['peak_sigma']
        delta_m    = mass_dict['delta_m']
        m2_min     = mass_dict['m2_min']
        m2_max     = mass_dict['m2_max']
        m1_m2      = 1

        m1_vec     = power_peak(m1_min, m1_max, m1_alpha, peak_frac, peak_mean, peak_sigma, delta_m, num_injs, nm1s=5001, rng=rngs[0])
        m2_vec     = rngs[1].uniform(low=m2_min, high=m2_max, size=num_injs)

    elif mass_dict['dist'] == 'power_uniform':
        mmin  = mass_dict['mmin']
        mmax  = mass_dict['mmax']
        alpha = mass_dict['alpha'] + 1
        m1_m2 = 1

        m1_vec = (mmin**alpha + (mmax**alpha - mmin**alpha)*rngs[0].random(num_injs))**(1./alpha)
        m2_vec = rngs[1].uniform(mmin,m1_vec)

    elif mass_dict['dist'] == 'uniform':
        mmin  = mass_dict['mmin']
        mmax  = mass_dict['mmax']
        m1_m2 = 0

        m1_vec = rngs[0].uniform(low=mmin, high=mmax, size=num_injs)
        m2_vec = rngs[1].uniform(low=mmin, high=mmax, size=num_injs)

    elif mass_dict['dist'] == 'fixed_m1_q':
        m1    = mass_dict['m1']
        if mass_dict['q'] > 1: m2 = m1 / mass_dict['q']
        else:                  m2 = m1 * mass_dict['q']
        m1_m2 = 1

        m1_vec = m1 * np.ones(num_injs)
        m2_vec = m2 * np.ones(num_injs)

    if m1_m2: return m1_vec, m2_vec
    else:     return make_m1_m2(m1_vec,m2_vec,num_injs!=1)

#-----mass handling functions-----
def get_Mc_eta(m1_vec,m2_vec):
    eta_vec = brs.eta_of_q(m1_vec/m2_vec)
    Mc_vec  = brs.Mc_of_M_eta(m1_vec+m2_vec,eta_vec)
    return Mc_vec, eta_vec

def make_m1_m2(m1,m2,vec=1):
    if vec:
        mt      = copy(m1)
        ids     = np.where(m1<m2)
        m1[ids] = m2[ids]
        m2[ids] = mt[ids]
        return m1, m2
    else:
        if m1 < m2: return m2, m1
        else:       return m1, m2

#-----power-peak helpers-----
def power(m, alpha):
    return m**(alpha)

def gaussian(m, mean, sigma):
    return np.exp(-((m - mean) / sigma)**2 / 2)

def smoothing(m,mmin,delta_m):
    m_arr  = np.array(m)
    res    = np.zeros_like(m_arr)
    res[np.nonzero(m_arr >= mmin + delta_m)[0]] = 1
    ids    = np.nonzero(np.logical_and(m_arr >= mmin, m_arr < mmin + delta_m))[0]
    m_arr -= mmin
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        res[ids] = 1 / (1 + np.exp(delta_m/m_arr[ids] + delta_m/(m_arr[ids] - delta_m)))
    return res

def power_peak(mmin, mmax, m1_alpha, peak_frac, peak_mean, peak_sigma, delta_m, num_injs, nm1s=5001, rng=None, seed=None):
    if rng is None: rng = np.random.default_rng(seed)
    m1s            = np.linspace(mmin, mmax, nm1s)
    power_part     = (1 - peak_frac) * power(m1s, -m1_alpha) / simps(power(m1s, -m1_alpha), m1s)
    gauss_part     = peak_frac * gaussian(m1s, peak_mean, peak_sigma) / simps(gaussian(m1s, peak_mean, peak_sigma), m1s)
    m1_dist        = (power_part + gauss_part) * smoothing(m1s, mmin, delta_m)
    window_cdf     = np.array([simps(m1_dist[:i], m1s[:i]) for i in range(1, len(m1s)+1)]) / simps(m1_dist, m1s)
    inv_window_cdf = interp1d(window_cdf, m1s)
    return inv_window_cdf(rng.random(num_injs))

#-----general helpers-----
def truncated_gaussian(mean, sigma, minv, maxv, num_injs, rng=None, seed=None):
    if rng is None: rng = np.random.default_rng(seed)
    sample_vec = np.zeros(num_injs)
    ids        = np.arange(num_injs)
    while ids.size > 0:
        sample_vec[ids] = rng.normal(loc=mean, scale=sigma, size=ids.size)
        ids             = np.nonzero(np.logical_not(np.logical_and(sample_vec > minv, sample_vec < maxv)))[0]
    return sample_vec


###
#-----redshift and lum distance samplers-----
def redshift_lum_distance_sampler(cosmo_dict,num_injs,seed):
    zmin = cosmo_dict['zmin']
    zmax = cosmo_dict['zmax']

    keys = list(cosmo_dict.keys())
    if 'Om0' in keys:  Om0 = cosmo_dict['Om0']
    else:              Om0 = None
    if 'Ode0' in keys: Ode0 = cosmo_dict['Ode0']
    else:              Ode0 = None
    if 'H0' in keys:   H0 = cosmo_dict['H0']
    else:              H0 = None

    if None in (Om0,Ode0,H0): cosmo = apcosm.Planck18
    else:                     cosmo = apcosm.LambdaCDM(H0=H0, Om0=Om0, Ode0=Ode0)

    if   cosmo_dict['sampler'] == 'uniform':
        z_vec = np.random.default_rng(seed).uniform(low=zmin, high=zmax, size=num_injs)
    elif cosmo_dict['sampler'] == 'uniform_in_bins':
    	z_vec = np.concatenate([
		    np.random.default_rng((i+1)*seed).uniform(low=zbin[0], high=zbin[1], size=cosmo_dict['injs_per_bin'][i])
		    for i,zbin in enumerate(cosmo_dict['bins']) ])
    elif cosmo_dict['sampler'] == 'uniform_comoving_volume_inversion':
        z_vec = uniform_comoving_volume_redshift_inversion_sampler(zmin,zmax,cosmo,num_injs,seed,nzs=None)
    elif cosmo_dict['sampler'] == 'uniform_comoving_volume_rejection':
        z_vec = uniform_comoving_volume_redshift_rejection_sampler(zmin,zmax,cosmo,num_injs,seed,nzs=40)
    elif cosmo_dict['sampler'] == 'mdbn_rate_inversion':
        z_vec = mdbn_merger_rate_uniform_comoving_volume_redshift_inversion_sampler(zmin,zmax,cosmo,num_injs,seed,nzs=None)
    elif cosmo_dict['sampler'] == 'bns_md_rate_inversion':
        z_vec = bns_md_merger_rate_uniform_comoving_volume_redshift_inversion_sampler(zmin,zmax,cosmo,num_injs,seed,nzs=None)
    elif cosmo_dict['sampler'] == 'popIII_rate_inversion':
        z_vec = popIII_merger_rate_uniform_comoving_volume_redshift_inversion_sampler(zmin, zmax, cosmo, num_injs, seed, nzs=None)
    elif cosmo_dict['sampler'] == 'pbh_power_age':
        z_vec = pbh_universe_age_power_law_sampler(zmin, zmax, cosmo, num_injs, seed)

    np.random.default_rng(seed).shuffle(z_vec)

    return z_vec, cosmo.luminosity_distance(z_vec).value

#-----redshift samplers-----
def uniform_comoving_volume_redshift_inversion_sampler(zmin,zmax,cosmo,num_injs,seed=None,nzs=None):
    rng = np.random.default_rng(seed)
    if nzs is None: nzs = max(5001, 50 * int((zmax-zmin)) + 1)
    zs = np.linspace(zmin,zmax,nzs)
    dist = (lambda z: ((4.*PI*cosmo.differential_comoving_volume(z).value)/(1.+z)))(zs)
    window_cdf = np.array([simps(dist[:i],zs[:i]) for i in range(1,nzs+1)]) / simps(dist,zs)
    inv_window_cdf = interp1d(window_cdf, zs)
    return inv_window_cdf(rng.random(num_injs))

def uniform_comoving_volume_redshift_rejection_sampler(zmin,zmax,cosmo,num_injs,seed=None,nzs=None):
    rng = np.random.default_rng(seed)
    dist = lambda z: ((4.*PI*cosmo.differential_comoving_volume(z).value)/(1.+z))
    window_norm = quad(dist, zmin, zmax)[0]
    flip_window_pdf = lambda z: -dist(z) / window_norm
    window_pdf_max = -minimize_scalar(flip_window_pdf,bounds=[zmin,zmax],method='bounded').fun

    if nzs is None or nzs < 2: nzs = 2
    zs = np.linspace(zmin,zmax,nzs)
    segment_nums = np.asarray((num_injs * np.array([-quad(flip_window_pdf,zs[i],zs[i+1])[0] for i in range(nzs-1)])), dtype=int)
    segment_pts = np.sum(segment_nums)
    segment_maxs = np.array([-minimize_scalar(flip_window_pdf,bounds=[zs[i],zs[i+1]],method='bounded').fun for i in range(nzs-1)])

    z_sample = np.zeros(num_injs)

    for j,num in enumerate(segment_nums):
        id_shift = np.sum(segment_nums[:j+1]) - num
        ids = np.arange(num) + id_shift
        while ids.size > 0:
            z_sample[ids] = rng.uniform(zs[j],zs[j+1],ids.size)
            ids = ids[np.nonzero(rng.uniform(0.,segment_maxs[j],ids.size) >= -flip_window_pdf(z_sample[ids]))[0]]

    if nzs > 2:
        ids = np.arange(num_injs - segment_pts) + segment_pts
        while ids.size > 0:
            z_sample[ids] = rng.uniform(zmin,zmax,ids.size)
            ids = ids[np.nonzero(rng.uniform(0.,window_pdf_max,ids.size) >= -flip_window_pdf(z_sample[ids]))[0]]
        rng.shuffle(z_sample)

    return z_sample

def mdbn_merger_rate_uniform_comoving_volume_redshift_inversion_sampler(zmin,zmax,cosmo,num_injs,seed=None,nzs=None):
    rng = np.random.default_rng(seed)
    if nzs is None: nzs = max(5001, 50 * int((zmax-zmin)) + 1)
    zs = np.linspace(zmin,zmax,nzs)
    dist = (lambda z: ((mdbn_merger_rate(z)*4.*PI*cosmo.differential_comoving_volume(z).value)/(1.+z)))(zs)
    window_cdf = np.array([simps(dist[:i],zs[:i]) for i in range(1,nzs+1)]) / simps(dist,zs)
    inv_window_cdf = interp1d(window_cdf, zs)
    return inv_window_cdf(rng.random(num_injs))

def bns_md_merger_rate_uniform_comoving_volume_redshift_inversion_sampler(zmin,zmax,cosmo,num_injs,seed=None,nzs=None):
    rng = np.random.default_rng(seed)
    if nzs is None: nzs = max(5001, 50 * int((zmax-zmin)) + 1)
    zs = np.linspace(zmin,zmax,nzs)
    dist = (lambda z: ((bns_md_merger_rate(z)*4.*PI*cosmo.differential_comoving_volume(z).value)/(1.+z)))(zs)
    window_cdf = np.array([simps(dist[:i],zs[:i]) for i in range(1,nzs+1)]) / simps(dist,zs)
    inv_window_cdf = interp1d(window_cdf, zs)
    return inv_window_cdf(rng.random(num_injs))

def popIII_merger_rate_uniform_comoving_volume_redshift_inversion_sampler(zmin, zmax, cosmo, num_injs, seed=None, nzs=None):
    rng            = np.random.default_rng(seed)
    if nzs is None: nzs = max(5001, 50 * int((zmax - zmin)) + 1)
    zs             = np.linspace(zmin, zmax, nzs)
    dist           = (lambda z: ((popIII_merger_rate(z) * 4. * PI * cosmo.differential_comoving_volume(z).value) / (1. + z)))(zs)
    window_cdf     = np.array([simps(dist[:i], zs[:i]) for i in range(1, nzs + 1)]) / simps(dist, zs)
    inv_window_cdf = interp1d(window_cdf, zs)
    return inv_window_cdf(rng.random(num_injs))

# PBH redshift distribution following a power law wrt the age of the universe (https://arxiv.org/pdf/2204.11864.pdf, Eq. (5))
def pbh_universe_age_power_law_sampler(zmin, zmax, cosmo, num_injs, seed):
    rng    = np.random.default_rng(seed)
    tau_0  = cosmo.age(0)
    x_min  = cosmo.age(zmin) / tau_0
    x_max  = cosmo.age(zmax) / tau_0
    x_vec  = rng.power(-34/37 + 1, num_injs)
    x_mask = np.logical_or(x_vec < x_max, x_vec > x_min)

    while x_mask.sum():
        x_vec[x_mask] = rng.power(-34/37 + 1, x_mask.sum())
        x_mask        = np.logical_or(x_vec < x_max, x_vec > x_min)

    x_vec  *= tau_0.value
    tau_vec = np.geomspace(np.sort(x_vec)[0], np.sort(x_vec)[-1], 100)
    return np.interp(x_vec, tau_vec, apcosm.z_at_value(cosmo.age, tau_vec * tau_0.unit).value)

#-----merger rate functions-----
# 'Madau-Dickinson-Belczynski-Ng' field BBH volumetric merger rate:
# https://arxiv.org/pdf/2012.09876.pdf, Eq. (C13) with F-values from page 13 (v3 of the paper)
def mdbn_merger_rate(z, a0=2.57, b0=5.83, c0=3.36, phi0=1):
    return phi0 * (1+z)**a0 / (1+((1+z)/c0)**b0)

# simple Madau-Dickinson star formation rate and 1/t time delay; metalicity was not taken into account
# fit of mdbn_merger_rate to the data found in 'xtra_files/merger_rates/bns_n_dot_bns_md_merger_rate.txt'
def bns_md_merger_rate(z, a0=1.803219571, b0=5.309821767, c0=2.837264101, phi0=8.765949529):
    return phi0 * (1+z)**a0 / (1+((1+z)/c0)**b0)

# BBH from popIII volumetric merger rate:
# https://arxiv.org/pdf/2012.09876.pdf, Eq. (C15) with III-values from page 13 (v3 of the paper)
def popIII_merger_rate(z, aIII=0.66, bIII=0.3, zIII=11.6):
    return np.exp(aIII * (z - zIII)) / (bIII + aIII * np.exp((aIII + bIII) * (z - zIII)))



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


"""This module handles the benchmarking of graviational waveforms observed by a network of detectors.

"""

import logging
from copy import copy, deepcopy

import dill
import numpy as np
from pathos.pools import ProcessPool

import gwbench.detector as dc
import gwbench.detector_response_derivatives as drd
import gwbench.err_deriv_handling as edh
import gwbench.fisher_analysis_tools as fat
import gwbench.waveform as wfc
import gwbench.utils as utils

# logger
glob_logger = logging.getLogger('network_module')
glob_logger.setLevel('INFO')

class Network:

    ###
    #-----Init methods-----
    def __init__(self, network_spec=None, logger_name='Network', logger_level='INFO'):
        ##-----logger-----
        self.logger = utils.get_logger(name=logger_name, level=logger_level)

        ##-----initialize network object-----
        if network_spec is None:
            #-----network and detectors
            # network label
            self.label = None
            # detector labels list
            self.det_keys = None
            # list of detectors in network
            self.detectors = None
        elif isinstance(network_spec, str):
            #-----network and detectors
            self.set_network_and_detectors_from_label(network_spec)
        elif isinstance(network_spec, list) or isinstance(network_spec, tuple):
            #-----network and detectors
            self.set_network_and_detectors_from_key_list(network_spec)

        #-----injection and waveform quantities-----
        # frequency array
        self.f = None
        # dictionary of injection parameters
        self.inj_params = None
        # derivative variables - symbs_string and list
        self.deriv_symbs_string = None
        self.deriv_variables = None
        # derivative variables for which analytical derivatives should be used
        self.ana_deriv_symbs_string = None
        self.ana_deriv_symbs_list = None
        # waveform
        self.wf = None

        #-----analysis settings-----
        # list of inj_params to convert to cos, ln versions
        self.conv_cos = None
        self.conv_log = None
        # use f-dependent gmst (SPA) in antenna patterns
        self.use_rot = None

        #-----user defined detector locations and PSDs-----
        self.user_locs = None
        self.user_psds = None
        self.user_lambdified_functions_path = None

        #-----waveform polarizations-----
        # plus/cross polarizations
        self.hfp = None
        self.hfc = None
        # derivative dictionary for polarizations
        self.del_hfpc = None
        # sympy expressions of derivative dictionary for polarizations
        self.del_hfpc_expr = None

        #-----network SNR-----
        # SNR, SNR^2 calculated from hf
        self.snr = None
        self.snr_sq = None

        #-----network errors-----
        # Fisher matrix
        self.fisher = None
        # condition number of Fisher matrix
        self.cond_num = None
        # covariance matrix
        self.cov = None
        # dictionary containing information about the inversion error between the two matrices
        self.inv_err = None
        # dictionary of errors for given derivative variables
        self.errs = None

        if self.label is None: self.logger.debug('Empty network initialized.')
        else:                  self.logger.debug('Network initialized.')


    ###
    #-----Setter methods-----
    #
    # it is best practice to always change the instance variables using these setter methods
    #
    def set_logger_level(self, level):
        utils.set_logger_level(self.logger, level)

    def set_network_and_detectors_from_label(self, network_label):
        #-----network and detectors
        self.label     = network_label
        self.det_keys  = utils.read_det_keys_from_label(network_label)
        self.detectors = [ dc.Detector(det_key) for det_key in self.det_keys ]

    def set_network_and_detectors_from_key_list(self, det_key_list):
        #-----network and detectors
        if isinstance(det_key_list, tuple): self.det_keys = [ tec + '_' + loc for tec,loc in zip(det_key_list[0], det_key_list[1]) ]
        else:                               self.det_keys = copy(det_key_list)
        self.label     = '..'.join(self.det_keys)
        self.detectors = [ dc.Detector(det_key) for det_key in self.det_keys ]

    def set_wf_vars(self, wf_model_name, wf_other_var_dic=None, user_waveform=None, cosmo=None):
        self.wf = wfc.Waveform(wf_model_name, wf_other_var_dic=wf_other_var_dic, user_waveform=user_waveform, cosmo=cosmo, logger=self.logger)

    def set_net_vars(self, f=None, inj_params=None, deriv_symbs_string=None, conv_cos=None, conv_log=None,
                     use_rot=None, user_locs=None, user_psds=None, user_lambdified_functions_path=None, ana_deriv_symbs_string=None):
        if f is not None:
            self.f = copy(f)
            if self.detectors is not None:
                for det in self.detectors:
                    det.set_f(self.f)
        if inj_params is not None:
            self.inj_params = deepcopy(inj_params)
        if deriv_symbs_string is not None:
            self.deriv_symbs_string = copy(deriv_symbs_string)
            self.deriv_variables = deriv_symbs_string.split(' ')
        if ana_deriv_symbs_string is not None:
            self.ana_deriv_symbs_string = copy(ana_deriv_symbs_string)
            self.ana_deriv_symbs_list = ana_deriv_symbs_string.split(' ')
        if conv_cos is not None:
            self.conv_cos = copy(conv_cos)
        if conv_log is not None:
            self.conv_log = copy(conv_log)
        if use_rot is not None:
            self.use_rot = copy(use_rot)
        if user_locs is not None:
            self.user_locs = deepcopy(user_locs)
        if user_psds is not None:
            self.user_psds = deepcopy(user_psds)
        if user_lambdified_functions_path is not None:
            self.user_lambdified_functions_path = copy(user_lambdified_functions_path)

    ##
    #-----Resetter methods for instance variables-----
    def reset_ant_pat_lpf_psds(self):
        for det in self.detectors:
            det.Fp = None
            det.Fc = None
            det.Flp = None
            det.psd = None

    def reset_wf_polarizations(self):
        self.hfp = None
        self.hfc = None
        self.del_hfpc = None
        self.del_hfpc_expr = None

    def reset_det_responses(self):
        for det in self.detectors:
            det.hf = None
            det.del_hf = None
            det.del_hf_expr = None

    def reset_snrs(self):
        self.snr = None
        self.snr_sq = None
        for det in self.detectors:
            det.snr = None
            det.snr_sq = None
            det.d_snr_sq = None

    def reset_errors(self):
        self.fisher = None
        self.cond_num = None
        self.cov = None
        self.inv_err = None
        self.errs = None
        for det in self.detectors:
            det.fisher = None
            det.cond_num = None
            det.cov = None
            det.inv_err = None
            det.errs = None


    ###
    #-----Getters-----
    def get_detector(self,det_key):
        return self.detectors[self.det_keys.index(det_key)]

    def get_snrs_errs_cov_fisher_inv_err_for_key(self,key='network'):
        if key == 'network': out_obj = self
        else:                out_obj = self.detectors[self.det_keys.index(key)]
        return out_obj.snr, out_obj.errs, out_obj.cov, out_obj.fisher, out_obj.inv_err


    ###
    #----Helper-----
    def pool_calc_func(self, calc_func, num_cores):
        if num_cores is None: list(map(calc_func, self.detectors))
        else:
            self.detectors, self.inj_params, self.deriv_variables = list(zip(*ProcessPool(nodes=num_cores).imap(calc_func, self.detectors)))
            self.inj_params                                       = self.inj_params[0]
            self.deriv_variables                                  = self.deriv_variables[0]


    ###
    #-----PSDs and antenna patterns-----
    def setup_ant_pat_lpf_psds(self, F_lo=-np.inf, F_hi=np.inf):
        self.setup_psds(F_lo, F_hi)
        self.setup_ant_pat_lpf()

    def setup_psds(self, F_lo=-np.inf, F_hi=np.inf):
        for det in self.detectors:
            det.setup_psds(F_lo, F_hi, user_psds=self.user_psds)
        self.logger.info('PSDs loaded.')

    def setup_ant_pat_lpf(self):
        for det in self.detectors:
            det.setup_ant_pat_lpf(self.inj_params, self.use_rot, user_locs=self.user_locs)
        self.logger.info('Antenna patterns and LPFs loaded.')


    ###
    #-----Waveform polarizations-----
    def calc_wf_polarizations(self):
        self.hfp, self.hfc = self.wf.eval_np_func(self.f,utils.get_sub_dict(self.inj_params,self.wf.wf_symbs_string))
        self.logger.info('Polarizations calculated.')

    def calc_wf_polarizations_derivs_num(self, step=1e-9, method='central', order=2, d_order_n=1):
        self.logger.info('Calculate numeric derivatives of polarizations.')
        self.calc_wf_polarizations()
        self.del_hfpc = drd.calc_det_responses_derivs_num(None, self.wf, utils.remove_symbols(self.deriv_symbs_string,self.wf.wf_symbs_string),
                                                          self.f, self.inj_params, use_rot=self.use_rot, label='hf',
                                                          step=step, method=method, order=order, d_order_n=d_order_n, user_locs=self.user_locs,
                                                          ana_deriv_symbs_string=self.ana_deriv_symbs_string, ana_deriv_aux={'hfp':self.hfp,'hfc':self.hfc})
        self.del_hfpc, c_quants = edh.get_conv_del_eval_dic(self.del_hfpc, self.inj_params, self.conv_cos, self.conv_log, self.deriv_symbs_string)
        self.inj_params, self.deriv_variables = edh.get_conv_inj_params_deriv_variables(c_quants, self.inj_params, self.deriv_variables)
        self.logger.info('Numeric derivatives of polarizations calculated.')

    def load_wf_polarizations_derivs_sym(self, return_bin=0):
        wf_deriv_symbs_string = utils.remove_symbols(self.deriv_symbs_string,self.wf.wf_symbs_string)
        self.del_hfpc_expr = drd.load_det_responses_derivs_sym('pl_cr', self.wf.wf_model_name, wf_deriv_symbs_string, self.use_rot, return_bin, self.user_lambdified_functions_path)
        self.logger.info('Lambdified polarizations loaded.')

    def calc_wf_polarizations_derivs_sym(self):
        self.logger.info('Evaluate polarizations.')
        self.calc_wf_polarizations()
        self.del_hfpc = { deriv : self.del_hfpc_expr[deriv](self.f, **utils.get_sub_dict(self.inj_params, self.del_hfpc_expr['variables']))
                          for deriv in self.del_hfpc_expr if deriv not in ('variables','deriv_variables')}
        self.del_hfpc, c_quants = edh.get_conv_del_eval_dic(self.del_hfpc, self.inj_params, self.conv_cos, self.conv_log, self.deriv_symbs_string)
        self.inj_params, self.deriv_variables = edh.get_conv_inj_params_deriv_variables(c_quants, self.inj_params, self.deriv_variables)
        self.logger.info('Lambdified polarizations evaluated.')


    ###
    #-----Detector responses-----
    def calc_det_responses(self):
        for det in self.detectors:
            det.calc_det_responses(self.wf,self.inj_params)
        self.logger.info('Detector responses calculated.')

    def calc_det_responses_derivs_num(self, step=1e-9, method='central', order=2, d_order_n=1, num_cores=None):
        self.logger.info('Calculate numeric derivatives of detector responses.')
        def calc_func(det):
            self.logger.info(f'   {det.det_key}')
            det.calc_det_responses_derivs_num(self.inj_params, self.deriv_variables, self.wf, self.deriv_symbs_string, self.conv_cos, self.conv_log,
                                              self.use_rot, step, method, order, d_order_n, self.user_locs, self.ana_deriv_symbs_string)
            return det, self.inj_params, self.deriv_variables
        self.pool_calc_func(calc_func, num_cores)
        self.logger.info('Numeric derivatives of detector responses calculated.')

    def load_det_responses_derivs_sym(self, return_bin=0):
        for det in self.detectors:
            det.load_det_responses_derivs_sym(self.wf.wf_model_name, self.deriv_symbs_string, self.use_rot, return_bin, self.user_lambdified_functions_path)
        self.logger.info('Lambdified detector responses loaded.')

    def calc_det_responses_derivs_sym(self, num_cores=None):
        self.logger.info('Evaluate lambdified detector responses.')
        def calc_func(det):
            self.logger.info(f'   {det.det_key}')
            det.calc_det_responses_derivs_sym(self.wf, self.inj_params, self.deriv_variables, self.conv_cos, self.conv_log, self.deriv_symbs_string)
            return det, self.inj_params, self.deriv_variables
        self.pool_calc_func(calc_func, num_cores)
        self.logger.info('Lambdified detector responses evaluated.')


    ###
    #-----SNR calculations-----
    def calc_snrs(self, only_net=0):
        self.snr_sq = 0
        for det in self.detectors:
             self.snr_sq += det.calc_snrs(only_net)
        self.snr = np.sqrt(self.snr_sq)
        self.logger.info('SNRs calculated.')

    def calc_snr_sq_integrand(self):
        for det in self.detectors:
            det.calc_snr_sq_integrand()
        self.logger.info('SNR integrands calculated.')


    ###
    #-----Error calculation and Fisher analysis-----
    def calc_errors(self, cond_sup=None, only_net=0):
        self.logger.info('Calculate errors (Fisher & cov matrices).')
        #-----calculate the error matrices: Fisher and Cov-----
        self.fisher = 0
        for det in self.detectors:
            self.logger.info(f'   {det.det_key}')
            self.fisher += det.calc_fisher_cov_matrices(only_net, cond_sup, logger=self.logger)
        self.cond_num = fat.calc_cond_number(self.fisher, logger=self.logger, tag='network')
        self.cov, self.inv_err = fat.calc_cov_inv_err(self.fisher, self.cond_num, cond_sup=cond_sup, logger=self.logger, tag='network')
        #-----calculate the absolute errors of the various variables-----
        self.errs = fat.get_errs_from_cov(self.cov,self.deriv_variables)
        if not only_net:
            for det in self.detectors:
                det.calc_errs(self.deriv_variables)
        self.logger.info('Errors calculated.')

    def calc_sky_area_90(self, only_net=0):
        if self.cov is None or self.errs is None:
            self.logger.warning("calc_sky_area_90: tag = network - Nothing done since either 'cov' or 'errs' are None.")
            return
        if 'ra' in self.deriv_variables and ('cos_dec' in self.deriv_variables or 'dec' in self.deriv_variables):
            if 'cos_dec' in self.deriv_variables: dec_str = 'cos_dec'
            else:                                 dec_str = 'dec'
            ra_id      = self.deriv_variables.index('ra')
            dec_id     = self.deriv_variables.index(dec_str)
            is_cos_dec = (dec_str == 'cos_dec')
            self.errs['sky_area_90'] = edh.sky_area_90(self.errs['ra'],self.errs[dec_str],self.cov[ra_id,dec_id],self.inj_params['dec'],is_cos_dec)
            if not only_net:
                for det in self.detectors:
                    det.calc_sky_area_90_network(ra_id, dec_id, self.inj_params['dec'], is_cos_dec, dec_str)
            self.logger.info('Sky areas calculated.')
        else:
            self.logger.warning('calc_sky_area_90: tag = network - Nothing done due to missing of either RA or COS_DEC (DEC) errors.')


    ###
    #-----IO methods-----
    def save_network(self,filename_path):
        '''Save the network under the given path using *dill*.'''
        with open(filename_path, "wb") as fi:
            dill.dump(self, fi, recurse=True)
        self.logger.info('Network pickled.')
        return

    def load_network(filename_path):
        '''Loading the network from the given path using *dill*.'''
        with open(filename_path, "rb") as fi:
            net = dill.load(fi)
        net.logger.info('Network loaded.')
        return net

    def print_network(self):
        sepl='--------------------------------------------------------------------------------------'
        print()
        print(sepl)
        print('Printing network.')
        print(sepl)
        print()
        for key,value in vars(self).items():
            if type(value) == dict:
                print('Key: ',key)
                for kkey in value.keys():
                    print('',kkey)
                    print('',value[kkey])
                print()
            elif value is not None:
                if key == 'wf':
                    print('Key: ',key)
                    for kkey,vvalue in vars(value).items():
                        print('',kkey.ljust(16,' '),'  ',vvalue)
                    print()
                else:
                    print('Key: ',key)
                    print(value)
                    print()
        print(sepl)
        print('Printing network done.')
        print(sepl)
        print()

    def print_detectors(self):
        sepl='--------------------------------------------------------------------------------------'
        sepl1='-------------------------------------------'
        print()
        print(sepl)
        print('Printing detectors.')
        print(sepl)
        for det in self.detectors:
            print(sepl1)
            print(det.det_key)
            print(sepl1)
            det.print_detector(0)
        print(sepl)
        print('Printing detectors done.')
        print(sepl)
        print()



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import lal
import lalsimulation as lalsim
from numpy import exp, pi

from gwbench.basic_relations import m1_m2_of_M_eta, M_of_Mc_eta, lam_12_of_lam_ts_eta
from gwbench.utils import Mpc, Msun

wf_symbs_string = 'f Mc eta chi1x chi1y chi1z chi2x chi2y chi2z DL tc phic iota lam_t delta_lam_t'

def hfpc(f, Mc, eta, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z, DL, tc, phic, iota, lam_t, delta_lam_t, approximant, is_lam12=0, fRef=0., phiRef=0.):
    f_min   = f[0]
    delta_f = f[1] - f[0]
    f_max   = f[-1] + delta_f

    if not fRef: fRef = f_min

    _m1, _m2 = m1_m2_of_M_eta(M_of_Mc_eta(Mc,eta),eta)
    _m1 *= Msun
    _m2 *= Msun
    _DL = DL * Mpc

    if is_lam12:
        # in this case: lam_t = lam1, delta_lam_t = lam2
        lam1 = lam_t
        lam2 = delta_lam_t
    else:
        lam1, lam2 = lam_12_of_lam_ts_eta(lam_t,delta_lam_t,eta)

    if 0 > lam1 or 0 > lam2: raise ValueError(f'Either lam1 ({lam1}) or lam2 ({lam2}) are smaller than 0.')

    tidal_params = lal.CreateDict();
    lalsim.SimInspiralWaveformParamsInsertTidalLambda1(tidal_params, lam1)
    lalsim.SimInspiralWaveformParamsInsertTidalLambda2(tidal_params, lam2)

    approx = lalsim.GetApproximantFromString(approximant)

    hPlus, hCross = lalsim.SimInspiralChooseFDWaveform(m1=_m1, m2=_m2,
                                   S1x = chi1x, S1y = chi1y, S1z = chi1z,
                                   S2x = chi2x, S2y = chi2y, S2z = chi2z,
                                   distance = _DL, inclination = iota, phiRef = phiRef,
                                   longAscNodes=0., eccentricity=0., meanPerAno = 0.,
                                   deltaF=delta_f, f_min=f_min, f_max=f_max, f_ref=fRef,
                                   LALpars=tidal_params, approximant=approx)

    pf = exp(1j*(2*f*pi*tc - phic))
    i0 = int(round((f_min-hPlus.f0) / delta_f))

    hfp = pf *  hPlus.data.data[i0:i0+len(f)]
    hfc = pf * hCross.data.data[i0:i0+len(f)]

    return hfp, hfc



# Copyright (C) 2020  Ssohrab Borhanian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import importlib.util
import os

import sympy as sp

from gwbench.utils import log_msg

from gwbench.wf_models import lal_bbh_np
from gwbench.wf_models import lal_bns_np
from gwbench.wf_models import tf2_np
from gwbench.wf_models import tf2_sp
from gwbench.wf_models import tf2_tidal_np
from gwbench.wf_models import tf2_tidal_sp

###
#-----Get waveform functions for np, sp and the symbols string based on the model name-----
def select_wf_model_quants(wf_model_name, user_waveform=None, cosmo=None, logger=None):

    if user_waveform is None:
        if wf_model_name == 'lal_bbh':
            np_mod = lal_bbh_np
            sp_mod = None
        elif wf_model_name == 'lal_bns':
            np_mod = lal_bns_np
            sp_mod = None
        elif wf_model_name == 'tf2':
            np_mod = tf2_np
            sp_mod = tf2_sp
        elif wf_model_name == 'tf2_tidal':
            np_mod = tf2_tidal_np
            sp_mod = tf2_tidal_sp
        else: log_msg(f'select_wf_model_quants: wf_model_name {wf_model_name} is not known!', level='ERROR')
    else:
        if 'np' in user_waveform: np_mod = load_module_from_file(user_waveform['np'])
        else: log_msg('select_wf_model_quants: user_waveform does not contain key = "np": numpy version of the waveform is needed!', level='ERROR')

        if 'sp' in user_waveform: sp_mod = load_module_from_file(user_waveform['sp'])
        else:                     sp_mod = None

    if sp_mod is None: sp_hfpc = None
    else:              sp_hfpc = sp_mod.hfpc

    return np_mod.wf_symbs_string, np_mod.hfpc, sp_hfpc

def load_module_from_file(file_path):
    spec   = importlib.util.spec_from_file_location(os.path.splitext(os.path.basename(file_path))[0], file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


class Waveform(object):

    ###
    #-----Init methods-----
    def __init__(self, wf_model_name=None, wf_other_var_dic=None, user_waveform=None, cosmo=None, logger=None):
        if wf_model_name is None:
            wf_symbs_string = None
            hfpc_np         = None
            hfpc_sp         = None
        else:
            wf_symbs_string, hfpc_np, hfpc_sp = \
                select_wf_model_quants(wf_model_name, user_waveform=user_waveform, cosmo=cosmo)

        self.wf_model_name    = wf_model_name
        self.wf_other_var_dic = wf_other_var_dic
        self.user_waveform    = user_waveform
        self.wf_symbs_string  = wf_symbs_string
        self.hfpc_np          = hfpc_np
        self.hfpc_sp          = hfpc_sp


    ###
    #-----Getter methods-----
    def get_sp_expr(self):
        if self.hfpc_sp is None: log_msg('get_sp_expr: Waveform does not have a sympy expression!', level='ERROR')

        symb_dic = {}
        for name in self.wf_symbs_string.split(' '):
            symb_dic[name] = sp.symbols(name,real=True)

        if self.wf_other_var_dic is None: return self.hfpc_sp(*list(symb_dic.values()))
        else:                             return self.hfpc_sp(*list(symb_dic.values()), *list(self.wf_other_var_dic.values()))

    def eval_np_func(self, f, inj_params):
        if isinstance(inj_params, dict):
            if self.wf_other_var_dic is None: return self.hfpc_np(f, **inj_params)
            else:                             return self.hfpc_np(f, **inj_params, **self.wf_other_var_dic)
        elif isinstance(inj_params, list):
            if self.wf_other_var_dic is None: return self.hfpc_np(f, *inj_params)
            else:                             return self.hfpc_np(f, *inj_params, **self.wf_other_var_dic)


    ###
    #-----IO methods-----
    def print_waveform(self):
        for key,value in vars(self).items():
            print(key.ljust(16,' '),'  ',value)
            print()
